/* draw a heightfield on the screen using libgrx (djgpp graphics) */
/* JPB 9/27/95 */

#include <grx.h>                /* djgpp libgrx graphics library */
#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <math.h>                       /* math stuff */
#include <string.h>                     /* strcpy() */
#include <conio.h>      /* non-portable! */
#include <ctype.h>                      /* tolower() */
#include "hcon.h"
#include "hf_io.h"

#define ESC '/033'

#define FALSE 0
#define TRUE 1
#define D double

struct Point3 {
 double x, y, z;                   /* three-dimensional point */
};

#define Elmod(vv, xq, yq)  vv[(((y10+yq) % ysize) * xsize) + \
			   ((x10 +xq) % xsize)]  /* elmt of array */

		/* draw two triangles joined along hypotenuse */
void tri3d(D x0, D z0,D x1,D z1,D y0,D y1,D y2,D y3, int s);
int hf_read(PTYPE **hh, unsigned int *xs, unsigned int *ys, char *fname);
void drawhf(PTYPE *d1, int maxx, int maxz, int s, int t);
int graph_setup(int mode, int *xsize, int *ysize);
void graph_clear();

D ex, ey, ez;      /* eye coordinate */
D zcenter;
D zfar, znear;
D zsf;      /* z coordinate scale factor */
D hs;      /* vertical (y) coordinate scale factor */
D ycorr;  /* y offset correction */
D zeps;   /* z epsilon */
int yoff;  /* screen coords y-offset */
int x10, y10;  /* hf size * 10 */
int maxcol = 255;  /* maximum color index */
short sxmid;  /* center coordinates of view window (eg 640x480) */
short symid;
struct Point3 light;  /* position of light source */

int main (int argc, char *argv[])
{
  PTYPE *d1;      /* height field array */
  int xsize, ysize;  /* heightfield size */
  int xgsize, ygsize; /* screen size */
  char c;
  D eyinc;
  D sq3;          /* constant, square-root of 3.0 */
  int shader;     /* 0 wireframe 1 height 2 slope 3 lightsource */
  int gmode = 3;      /* 0 320x200 1 640x480 2:800x600 3:1024x768 */
/* char *buf; */
  char *msg1 = "+/-:height  B/F:eye-horiz  U/D:eye-vert  S:shader  T:tile";
  char *msg2 = "M: toggle mode  H:toggle help   Q:quit";
  GrTextOption opt;
  int help = 1;
  int tile = 0;
  char infile[80] = "output.tga";

 
  if (argc > 1) {
    strcpy(infile,argv[1]);
  }

  if (readn_hf(&d1, &xsize, &ysize, infile)) {
      fprintf(stderr,"Could not read %s.\n",infile);
      exit(1);
    }
  x10 = 10 * xsize;
  y10 = 10 * ysize;
		
		
		/* set up graphics mode */
  if (graph_setup(gmode,&xgsize,&ygsize)) {
	  fprintf(stderr,"Error setting graphics mode.\n");
	  exit(1);
  }
  sxmid = xgsize / 2;  /* middle of screen in screen coords */
  symid = ygsize / 2;
  zsf = 0.3;      /* z coordinate scale factor */
  hs = -0.45;      /* vertical (y) coordinate scale factor */
  ycorr = 0.0;  /* y offset correction */
  zeps = 0.001;


	/* viewpoint 0,0,0 in +z direction */

  ex = 0.0; ey = -2; ez = -3;
  sq3 = 1.0/sqrt(3.0);
  light.x = -sq3; light.y = sq3; light.z = sq3;   /* direction vector */
  eyinc = 0.8;
  zcenter = 1;
  shader = 3;

  opt.txo_font = NULL;
  opt.txo_xalign = GR_ALIGN_LEFT;
  opt.txo_yalign = GR_ALIGN_TOP;
  opt.txo_xmag = 1;
  opt.txo_ymag = 1;
  opt.txo_fgcolor.v = 255;
  opt.txo_bgcolor.v = 0;
  opt.txo_direct = GR_TEXT_RIGHT;
  opt.txo_chrtype = GR_BYTE_TEXT;
  
  do {
    GrClearScreen(GrBlack());
    yoff = (int) (symid*(ycorr + ey/zcenter)); 
    drawhf(d1,xsize,ysize,shader,tile);
    if (help) {    
      GrDrawString(msg1,strlen(msg1),10,ysize/10,&opt);
      GrDrawString(msg2,strlen(msg2),10,ysize/10+GrFontHeight(&opt),&opt);
    }
    c = tolower(getch());
    
    if (c == '+') hs *= 1.3;
    if (c == '-') hs /= 1.3;
    if (c == 'u') ey -= eyinc;
    if (c == 'd') ey += eyinc;
    if (c == 'b') ez -= 1.0;
    if (c == 'f') {
      ez += 1.0;
      if (ez >= -1.1) ez = -1.1;
    }
    if (c == 'h') help = !(help);
    if (c == 't') tile = !(tile);
    if (c == 's') {
     shader++;
     if (shader > 3) shader = 0;
    }
    if (c == 'm') {
     gmode++;
     if (gmode > 3) gmode = 0;
     if (graph_setup(gmode,&xgsize,&ygsize)) {
	  fprintf(stderr,"Error setting graphics mode.\n");
	  exit(1);
     }
     sxmid = xgsize / 2;  /* middle of screen in screen coords */
     symid = ygsize / 2;
    }

    zcenter = -ez;
    ycorr = 2*ey/(zcenter);
  } while (c != 'q');
  
  graph_clear();
  return(0);
}

int graph_setup(int mode, int *xsize, int *ysize)        
{
int i;
int red,green,blue;

/*        atype = GrAdapterType();
	  printf("Graphics adaper type = %d\n",atype);
	  mtype = getch(); */
	
	if (mode == 0)
	  GrSetMode(GR_width_height_color_graphics,320,200,256);
	else if (mode == 1)
	  GrSetMode(GR_width_height_color_graphics,640,480,256);
	else if (mode == 2)
	  GrSetMode(GR_width_height_color_graphics,800,600,256);
	else   /* mode == 3 */
	  GrSetMode(GR_width_height_color_graphics,1024,768,256);

	xsize[0] = GrSizeX();
	ysize[0] = GrSizeY();

	GrResetColors();
	for (i=0;i<256;i++) {
	  red = i;
	  green = i+3;
	  if (green > 255) green = 255;
	  blue = i;
	  GrSetColor(i,red,green,blue);
	}
    return (0);
} /* end graph_setup */

void graph_clear()
{
  GrSetMode(GR_default_text); 
}

/* =============================================================== */
/* tri3d() --
 * project two triangles, given 3D coordinates, on a 2D viewscreen
 * uses gobal vars:
 *   hs : vertical heightfield scaling factor
 *   ex,ey,ez : location of eye (viewpoint)
 *   symid : y coordinate midpoint
 *   ycorr : y correction factor
 *   yoff  : y offset in screen units 
 *   zcenter : center to 'rotate' appparent hf around
 *   maxcol  : maximum color value 
 * ===============================================================
 */

/*  shader: 0 = wireframe, 1 = height color, 2 = slope, 3 = lightsource */

void tri3d(D x0, D z0,D x1,D z1,D y0,D y1,D y2,D y3, int shader)
{
 D zmod0, zmod1; 
 D slope1, slope2;
 int c1, c2;  /* colors */
 D nx,ny,nz,nmag;   /* normal vector */
 D ctheta;          /* cos(theta)  theta = angle between light and normal */
 int p[3][2];       /* three sets of points in 2D */
 
 c1 = c2 = 0;       /* make compiler happy */

 if (shader == 1) {  /* height, before transform */
   c1 = (maxcol>>2)*(y0 + y1 + y2 + y3);
   c2 = c1;
 } 
 
/*
 printf("%f %f %f %f %f %f %f %f\n",x0,z0,x1,z1,y0,y1,y2,y3);
 getch(); */
 if (shader == 2) {
   slope1 = atan2( (z0-z1)*(x1-x0),(z1-z0)*(y2-y0) );
   c1 = (int) ((maxcol>>1)*(0.5*slope1 + 1.8));
   if (c1 < 0) c1 = 0;
   if (c1 > maxcol) c1 = maxcol;
  
   slope2 = atan2( (z0-z1)*(x1-x0),(z1-z0)*(y1-y3) ); 
   c2 = (int) ((maxcol>>1)*(0.5*slope2 + 1.8));
   if (c2 < 0) c2 = 0;
   if (c2 > maxcol) c2 = maxcol;
  
 } else if (shader == 3) { /* compute normal, dot product with lightsource */
   nx = (z0-z1)*(y2-y0);
   ny = (z1-z0)*(x1-x0);
   nz = (x1-x0)*(y2-y1);
   nmag = sqrt(nx*nx + ny*ny + nz*nz);
   ctheta = (nx*light.x + ny*light.y + nz*light.z) / nmag;
   c1 = (int) ((maxcol>>1)*(0.5*ctheta + 0.5));
   
   nx = -(z0-z1)*(y3-y1);
   /* ny = (z1-z0)*(x1-x0); */   /* duplicates line above */
   nz = (x0-x1)*(y3-y0);
   nmag = sqrt(nx*nx + ny*ny + nz*nz);
   ctheta = (nx*light.x + ny*light.y + nz*light.z) / nmag;
   c2 = (int) ((maxcol>>1)*(0.5*ctheta + 0.5));

 } else if (shader == 0) {  /* shader = 0: wireframe */
   c1 = maxcol;  /* white, we hope. brightest color, anyway */
   c2 = maxcol;
 }
 /* c2 = c1; */
 
 /* x0,y0,z0 should be in the range -1..1 */
 /* transform to eye coordinate (ex,ey,ez) */
 
 y0 *= hs; y1 *= hs; y2 *= hs; y3 *= hs;                          
 x0 -= ex; y0 -= ey; z0 -= ez;
 x1 -= ex; y1 -= ey; z1 -= ez;
 y2 -= ey; y3 -= ey; 
 if ((z1 <= zeps) || (z0 <= zeps)) return;  /* foreground clipping */


 zmod0 = z0 * zsf;   /* z1 multiplied by z scale factor */
 zmod1 = z1 * zsf;
/*
 printf("zmod: %f %f\n",zmod0,zmod1);
  getch(); */
  /* sxmid, symid are midpoints of screen, eg sxmid=320 for 640x480 scn. */

 p[0][0] = (int) (sxmid*x0/zmod0) + sxmid;  /* x,y,z 0 */
 p[0][1] = (int) (symid*y0/zmod0) + symid + yoff;
 
 p[1][0] = (int) (sxmid*x1/zmod1) + sxmid;  /* x,y,z 1 */
 p[1][1] = (int) (symid*y1/zmod1) + symid + yoff;

 p[2][0] = (int) (sxmid*x1/zmod0) + sxmid; /* x,y,z 2 */
 p[2][1] = (int) (symid*y2/zmod0) + symid + yoff;
 
 if( ((p[0][0]==p[1][0]) && (p[0][1]==p[1][1])) ||
     ((p[0][0]==p[2][0]) && (p[0][1]==p[2][1])) ||
     ((p[1][0]==p[2][0]) && (p[1][1]==p[2][1]))  )  {
       GrPlot(p[0][0],p[0][1],c1);
   } else {
     if (shader == 0) 
       GrPolygon(3,p,c1);
     else if (shader == 1) {
       GrFilledConvexPolygon(3,p,c1);   
       GrPolygon(3,p,128+c1/2);
     } else 
       GrFilledConvexPolygon(3,p,c1);   
   }

 p[1][0] = (int) (sxmid*x0/zmod0) + sxmid;  /* x,y,z 0 */
 p[1][1] = (int) (symid*y0/zmod0) + symid + yoff;

 p[0][0] = (int) (sxmid*x1/zmod1) + sxmid;  /* x,y,z 1 */
 p[0][1] = (int) (symid*y1/zmod1) + symid + yoff;

 p[2][0] = (int) (sxmid*x0/zmod1) + sxmid; /* x,y,z 3 */
 p[2][1] = (int) (symid*y3/zmod1) + symid + yoff; 

 if( ((p[0][0]==p[1][0]) && (p[0][1]==p[1][1])) ||
     ((p[0][0]==p[2][0]) && (p[0][1]==p[2][1])) ||
     ((p[1][0]==p[2][0]) && (p[1][1]==p[2][1]))  )  {
       GrPlot(p[0][0],p[0][1],c2);
   } else {
     if (shader == 0) GrPolygon(3,p,c2);
     else if (shader == 1) {
       GrFilledConvexPolygon(3,p,c2);   
       GrPolygon(3,p,128+c1/2);
     } else GrFilledConvexPolygon(3,p,c2);   
   }

} /* end tri3d() */

/* draw a heightfield, back to front, as a series of triangles */
void drawhf(PTYPE *d1, int xsize, int ysize, int shader, int tile)
{
  int ix,iz;
  int ymin, ymax, xmin, xmax;
  D xd2, yd2;
  D x0, y0, z0, x1, y1, z1;
  D y2, y3;
  
  xd2 = xsize / 2.0;
  yd2 = ysize / 2.0;
  if (tile) {
    ymin = 0;
    ymax = (2*ysize) - 1;
    xmin = -xsize + 1;
    xmax = 2*xsize -1;
  } else {
    ymin = 0;
    ymax = ysize - 2;
    xmin = 0;
    xmax = xsize-1;
  }
		     /* loop over the two array indices */        
    for ( iz=ymax;iz>=ymin;iz--) {
	  for ( ix=xmin;ix<xmax;ix++) {
	    x0 = (ix-xd2)/xd2; z0 = (iz-yd2)/yd2; 
	    x1 = (ix+1-xd2)/xd2; z1 = (iz+1-yd2)/yd2; 
	    y0 = Elmod(d1,ix,iz);
/*           printf("%d %d: %f\n",ix,iz,y0); */
	    y1 = Elmod(d1,ix+1,iz+1);
	    y2 = Elmod(d1,ix+1,iz);
	    y3 = Elmod(d1,ix,iz+1);
	    tri3d(x0,z0,x1,z1,y0,y1,y2,y3,shader); 
	    /* getch(); */
	  } /* end for iz */
    } /* end for ix */
} /* end drawhf() */


